
# from negf import pole_summation_method
#
#
# def test_density_integration():
#     """ """
#     ans = pole_summation_method()
#     ans1 = numerical_intrgration()
#     np.testing.assert_allclose(ans, ans1, rtol=1e-2)
